from .operators import OP_MAP as _OP_MAP

from .lib import *
from .lib import (
    _CAST_POLICIES,
    _itype_le, _itype_lt,
    _throw_MixedItype_err_or_warn,
    _find_least_common_itype,
)

import pandas as pd
import numpy as np
import operator as op

import functools as ftools

import pandas.core.dtypes.common as dcom

_is_list_like = dcom.is_list_like
_is_nested_list_like = dcom.is_nested_list_like
_is_scalar = dcom.is_scalar
_is_integer = dcom.is_integer
_is_dict_like = dcom.is_dict_like
_is_number = dcom.is_number
_is_hashable = dcom.is_hashable

from pandas.core.common import is_bool_indexer as _is_bool_indexer

"""
Unlike the example says, return lists False, not True
>>is_iterator([1, 2, 3])
>>False
"""
from pandas.core.dtypes.common import is_iterator as _is_iterator

from typing import Union, Any


class DictOfSeries:
    """ A data frame where every column has its own index.

    DictOfSeries is a collection of pd.Series's which aim to be as close as possible similar to
    pd.DataFrame. The advantage over pd.DataFrame is, that every `column` has its own row-index,
    unlike the former, which provide a single row-index for all columns. This solves problems with
    unaligned data and data which varies widely in length.

    Indexing with ``di[]``, ``di.loc[]`` and ``di.iloc[]``  should work analogous to these methods 
    from pd.DataFrame. The indexer can be a single label, a slice, a list-like, a boolean list-like, 
    or a boolean dios/pd.DataFrame and can be used to selectively get or set data.
    
    Parameters
    ----------
    data : array-like, Iterable, dict, or scalar value
        Contains data stored in Series.

    columns : array-like 
        Column labels to use for resulting frame. Will default to
        RangeIndex (0, 1, 2, ..., n) if no column labels are provided.
        
    itype : Itype, pd.Index, Itype-string-repr, type
        Index type that every series in this dios should have. 
        if None, the index-type is inferred each time a series is inserted
        or deleted.
        
    cast_policy : str
        Policy to use for down-casting an itype.
    """

    def __init__(self, data=None, columns=None, index=None, itype=None, cast_policy='save', fastpath=False):

        self.cast_policy = cast_policy

        # we are called internally
        if fastpath:
            self._itype = itype or ObjItype
            if data is not None:
                self._data = data
            else:
                # it is significantly faster, to provide an index and fill it,
                # than to successively build the index by adding data
                self._data = pd.Series(dtype='O', index=columns)

        else:

            if index is not None and not isinstance(index, pd.Index):
                index = pd.Index(index)

            # itype=None means infer the itype by the data, so we first set to the highest
            # possible itype, then insert data, then infer the best-fitting itype.
            if itype is None and index is None:
                self._itype = ObjItype
            else:
                if index is not None:
                    self._itype = get_itype(index)
                if itype is not None:
                    self._itype = get_itype(itype)

            cols = pd.Index([] if columns is None else columns)
            if not cols.is_unique:
                raise ValueError("columns must be unique")
            self._data = pd.Series(dtype='O', index=cols)

            if data is not None:
                self._init_insert_data(data, columns, index)

        # self._data still contain nans at all positions, where
        # no data was present, but a column-name was given
        if self._data.hasnans:
            e = pd.Series(dtype='O', index=index)
            for c in self.columns[self._data.isna()]:
                self._insert(c, e.copy())

        self._data.index.name = 'columns'

        # we try to infer the itype, but if we still have
        # no data, we will set the itype lazy, i.e. with
        # the first non-empty _insert()
        if itype is None:
            if self.empty:
                self._itype = 'INFER'
            else:
                self._itype = _find_least_common_itype(self._data)
                if not self._itype.unique:
                    _throw_MixedItype_err_or_warn(self.itype)

    def _init_insert_data(self, data, columns, index):
        """ Insert items of a iterable in self"""

        if _is_iterator(data):
            data = list(data)

        if _is_dios_like(data) or isinstance(data, dict):
            if columns is None:
                pass  # data is dict-like
            else:
                data = {k: data[k] for k in data if k in columns}

        elif isinstance(data, pd.Series):
            name = data.name or 0
            if columns is not None and len(columns) > 0:
                name = self.columns[0]
            data = {name: data}

        elif _is_nested_list_like(data):
            if columns is None:
                data = {i: d for i, d in enumerate(data)}
            elif len(data) == len(columns):
                data = dict(zip(self.columns, data))
            else:
                raise ValueError(f"{len(columns)} columns passed, data implies {len(data)} columns")

        elif _is_list_like(data):
            name = 0 if columns is None or len(columns) < 1 else self.columns[0]
            data = {name: data}

        else:
            raise TypeError("data type not understood")

        for k in data:
            self._insert(k, pd.Series(data[k], index=index))

    # ----------------------------------------------------------------------
    #

    def _insert(self, col, val):
        """Insert a fresh new value as pd.Series into self"""
        val = list(val) if _is_iterator(val) else val

        if _is_dios_like(val):
            val = val.squeeze()
            if not isinstance(val, pd.Series):
                raise ValueError(f"Cannot insert frame-like with more than one column")

        elif val is None:
            val = pd.Series()

        elif not isinstance(val, pd.Series):
            raise TypeError(f"Only data of type pandas.Series can be inserted, passed was {type(val)}")

        # set the itype lazy, i.e. when first non-empty
        # column is inserted
        if self._itype == 'INFER':
            if not val.empty:
                self._itype = get_itype(val.index)
                # cast all pre-inserted empty series
                self.__cast_all(self._itype, self._policy)
                if not self._itype.unique:
                    _throw_MixedItype_err_or_warn(self._itype)
        else:
            val = cast_to_itype(val, self.itype, policy=self._policy)

        val.name = col
        self._data.at[col] = val.copy(deep=True)

    @property
    def columns(self):
        return self._data.index

    @columns.setter
    def columns(self, cols):
        self._data.index = cols
        if not self._data.index.is_unique:
            raise ValueError("columns index must have unique values")

    @property
    def values(self):
        s = self._fast_apply('values')
        return s.values

    @property
    def indexes(self):
        """ Return pd.series with the indexes. """
        return self._fast_apply('index')

    def index_of(self, method='all'):
        """ Return an single index with indices from all columns.

        parameter
        ----------
        method: string
            - 'all' : get all indices from all columns
            - 'union' : alias for 'all'
            - 'shared' : get indices that are present in every columns
            - 'intersection' : alias for 'shared'
            - 'uniques' : get indices that are only present in a single column
            - 'non-uniques' : get indices that are present in more than one column

        Note
        ----
        The returned index never contains duplicates.
        """

        indexes = self.indexes
        if len(indexes) <= 1:
            return indexes.squeeze()

        if method in ['union', 'all']:
            res = ftools.reduce(pd.Index.union, indexes)
        elif method in ['intersection', 'shared']:
            res = ftools.reduce(pd.Index.intersection, indexes)
        elif method in ['uniques', 'non-uniques']:
            res = ftools.reduce(pd.Index.append, indexes)
            res = res.value_counts(sort=False, dropna=False)
            if method == 'uniques':
                res = res[res == 1].index
            else:
                res = res[res > 1].index
        else:
            raise ValueError(method)
        return res if res.is_unique else res.unique()

    @property
    def cast_policy(self):
        return self._policy

    @cast_policy.setter
    def cast_policy(self, policy):
        if policy not in _CAST_POLICIES:
            raise ValueError(f"policy must be one of {_CAST_POLICIES}")
        self._policy = policy

    @property
    def itype(self):
        if self._itype == 'INFER':
            return None
        return self._itype

    @itype.setter
    def itype(self, itype):
        itype = get_itype(itype)
        self.__cast_all(itype, policy=self._policy)
        self._itype = itype

    def __cast_all(self, itype, policy):
        c = '?'
        try:
            for c in self.columns:
                self[c] = cast_to_itype(self[c], itype, policy=policy)
        except Exception as e:
            raise type(e)(f"Column {c}: " + str(e)) from e

    def __getitem__(self, key):
        """ dios[key] -> dios/series """
        key = list(key) if _is_iterator(key) else key
        if isinstance(key, tuple):
            raise KeyError("tuples are not allowed")

        if _is_hashable(key):
            # NOTE: we use copy here to prevent index
            # changes, that could result in an invalid
            # itype. A shallow copy is not sufficient.

            # work on columns, return series
            return self._data.at[key].copy()

        if _is_dios_like(key):
            # work on rows and columns
            new = self._getitem_bool_dios(key)
        elif isinstance(key, slice):
            # work on rows
            new = self._slice(key)
        elif _is_bool_indexer(key):
            # work on rows
            new = self._getitem_bool_listlike(key)
        else:
            # work on columns
            data = self._data.loc[key]
            new = DictOfSeries(data=data, itype=self.itype, cast_policy=self._policy, fastpath=True)

        return new

    def _slice(self, key):
        """slices self, return copy"""
        if key == slice(None):
            return self.copy()

        new = self.copy_empty(columns=True)
        for k in self.columns:
            new._data.at[k] = self._data.at[k][key]
        return new

    def _getitem_bool_dios(self, key):
        """ Select items by a boolean dios-like drop un-selected indices. """

        if not _is_bool_dios_like(key):
            raise ValueError("Must pass DictOfSeries with boolean values only")

        new = self.copy_empty(columns=True)
        for k in self.columns.intersection(key.columns):
            dat = self._data.at[k]
            val = key[k]
            # align rows
            idx = val[val].index.intersection(dat.index)
            new._data.at[k] = dat[idx]
        return new

    def _getitem_bool_listlike(self, key):
        new = self.copy_empty(columns=True)
        for k in self.columns:
            new._data.at[k] = self._data.at[k].loc[key]
        return new

    def __setitem__(self, key, value):
        """ dios[key] = value """
        key = list(key) if _is_iterator(key) else key
        if isinstance(key, tuple):
            raise KeyError(f"{key}. tuples are not allowed")

        elif _is_hashable(key):
            if isinstance(value, pd.Series) or key not in self.columns:
                self._insert(key, value)
            elif _is_dios_like(value) or _is_nested_list_like(value):
                raise ValueError("Incompatible indexer with multi-dimensional value")
            else:
                self._data.at[key][:] = value

        else:
            data = self.__getitem__(key)
            assert isinstance(data, self.__class__), f"getitem returned data of type {type(data)}"

            # special cases
            if _is_dios_like(value):
                self._setitem_dios(data, value)
            # NOTE: pd.Series also considered list-like
            elif _is_list_like(value):
                self._setitem_listlike(data, value)

            # default case
            else:
                for k in data.columns:
                    s = data._data.at[k]
                    s[:] = value
                    self._data.at[k][s.index] = s

    def _setitem_listlike(self, data, value):

        value = value.values if isinstance(value, pd.Series) else value

        if len(value) != len(data.columns):
            raise ValueError(f"array-like value of length {len(value)} could "
                             f"not be broadcast to indexing result of shape "
                             f"(.., {len(data.columns)})")

        for i, k in enumerate(data.columns):
            s = data._data.at[k]
            s[:] = value[i]
            self._data.at[k][s.index] = s

    def _setitem_dios(self, data, value):
        """ Write values from a dios-like to self.

        No justification or alignment of columns, but of indices.
        If value has missing indices, nan's are inserted at that
        locations, just like `series.loc[:]=val` or `df[:]=val` do.

        Eg.
         di[::2] = di[::3]   ->   di[::2]

            x |        x |            x |
        ===== |     ==== |       ====== |
        0   x |     0  z |       0    z |
        2   x |  =  3  z |   ->  2  NaN |
        4   x |     6  z |       4  NaN |
        6   x |                  6    z |

        Parameter
        ----------
        data : dios
            A maybe trimmed version of self
        value : dios, pd.Dataframe
            The value to set with the same column dimension like data
        """

        if len(data) != len(value.columns):
            raise ValueError(f"shape mismatch: values array of shape "
                             f"(.., {len(value.columns)}) could not "
                             f"be broadcast to indexing result of "
                             f"shape (.., {len(data.columns)})")

        for i, k in enumerate(data):
            dat = data._data.at[k]
            # .loc cannot handle empty series,
            # like `emptySeries.loc[:] = [1,2]`
            if dat.empty:
                continue
            val = value[value.columns[i]]
            dat.loc[:] = val
            self._data.at[k].loc[dat.index] = dat

    @property
    def loc(self):
        return _LocIndexer(self)

    @property
    def iloc(self):
        return _iLocIndexer(self)

    @property
    def aloc(self):
        return _aLocIndexer(self)

    @property
    def at(self):
        return _AtIndexer(self)

    @property
    def iat(self):
        return _iAtIndexer(self)

    @property
    def empty(self):
        return len(self) == 0 or all(s.empty for s in self._data)

    def equals(self, other):
        try:
            eqna = (self.isna() == other.isna()).all(None)
            return eqna and (self.dropna() == other.dropna()).all(None)
        except Exception:
            return False

    def _reduce_horizontal(self, function, initializer_value):
        res = pd.Series(data=initializer_value, index=self.index_of('all'))
        for d in self._data:
            base = res.loc[d.index]
            if len(base) > 0:
                res.loc[d.index] = function(base, d)
        return res

    def all(self, axis=0):
        if axis in [0, 'index']:
            return self._data.apply(all)
        elif axis in [1, 'columns']:
            func = lambda s1, s2: s1.astype(bool) & s2.astype(bool)
            return self._reduce_horizontal(func, True)
        elif axis is None:
            return self._data.apply(all).all()
        raise ValueError(axis)

    def any(self, axis=0):
        if axis in [0, 'index']:
            return self._data.apply(any)
        elif axis in [1, 'columns']:
            func = lambda s1, s2: s1.astype(bool) | s2.astype(bool)
            return self._reduce_horizontal(func, False)
        elif axis is None:
            return self._data.apply(any).any()
        raise ValueError(axis)

    def squeeze(self, axis=None):
        if axis in [0, 'index']:
            if (self.lengths == 1).all():
                return self._data.apply(pd.Series.squeeze)
            return self
        elif axis in [1, 'columns']:
            if len(self) == 1:
                return self._data.squeeze()
            return self
        elif axis is None:
            if len(self) == 1:
                return self._data.squeeze().squeeze()
            if (self.lengths == 1).all():
                return self._data.apply(pd.Series.squeeze).squeeze()
            return self
        raise ValueError(axis)

    @property
    def size(self):
        return self.lengths.sum()

    @property
    def lengths(self):
        return self._data.apply(len)

    def __len__(self):
        return len(self.columns)

    def __iter__(self):
        yield from self.columns

    def __reversed__(self):
        yield from reversed(self.columns)

    def __contains__(self, item):
        return item in self.columns

    def __delitem__(self, key):
        del self._data[key]

    def __copy__(self):
        return self.copy(deep=True)

    def __deepcopy__(self, memo=None):
        return self.copy(deep=True)

    def copy(self, deep=True):
        if deep:
            data = pd.Series(dtype='O', index=self.columns)
            for c in self.columns:
                data.at[c] = self._data.at[c].copy(deep=True)
        else:
            data = self._data
        kws = dict(itype=self._itype, cast_policy=self._policy)
        return DictOfSeries(data=data, fastpath=True, **kws)

    def copy_empty(self, columns=True):
        data = None
        if columns is True:  # is correct
            data = pd.Series(dtype='O', index=self.columns)
            for c in self.columns:
                data.at[c] = pd.Series(dtype=self._data.at[c].dtype)
        kws = dict(itype=self._itype, cast_policy=self._policy)
        return DictOfSeries(data=data, fastpath=True, **kws)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        repr = dios_options[OptsFields.dios_repr]
        showdim = self.lengths.max() > dios_options[OptsFields.disp_max_rows]
        return self.to_string(method=repr, show_dimensions=showdim)

    def to_string(self, max_rows=None, min_rows=None, max_cols=None,
                  na_rep='NaN', show_dimensions=False,
                  method=Opts.repr_indexed,
                  no_value=' ', empty_series_rep='no data',
                  col_delim=' | ', header_delim='=', col_space=None, ):
        """ Pretty print a dios.

        if `method` == `indexed` (default):
          every column is represented by a own index and corresponding values

        if `method` == `aligned` [2]:
          one(!) global index is generated and values from a column appear at
          the corresponding index-location.

        Common params :
         - max_cols: not more column than `max_cols` are printed [1]
         - max_rows: see `min_rows` [1]
         - min_rows: not more rows than `min_rows` are printed, if rows of any series exceed `max_rows` [1]
         - na_rep: all NaN-values are replaced by `na_rep`. Default `NaN`

        Indexed-Method params:
         - empty_series_rep: empty series are represented by the string in `empty_series_rep`
         - col_delim: between all columns `col_delim` is inserted. (str)
         - header_delim: between the column names (header) and the data, `header_delim` is inserted,
           if not None. The string is repeated, up to the width of the column. (str or None).

        Aligned-Method params:
        - no_value: value that indicates, that no entry in the underling series is present. Bear in mind
          that this should differ from `na_rep`, otherwise you cannot differ missing- from NaN- values.

        Notes:
          [1]: defaults to the corresponding value in `dios_options`
          [2]: the common-params are directly passed to pd.DataFrame.to_string(..)
               under the hood, if method is `aligned`

        """
        if self.empty:
            return _empty_repr(self)

        max_cols = max_cols or dios_options[OptsFields.disp_max_cols] or 100
        max_rows = max_rows or dios_options[OptsFields.disp_max_rows] or 200
        min_rows = min_rows or dios_options[OptsFields.disp_min_rows] or 100

        kwargs = dict(max_rows=max_rows, min_rows=min_rows, max_cols=max_cols,
                      na_rep=na_rep, col_space=col_space, show_dimensions=show_dimensions)

        if method == Opts.repr_aligned:
            return _to_aligned_df(self, no_value=no_value).to_string(**kwargs)

        # add pprint relevant options
        kwargs.update(empty_series_rep=empty_series_rep,
                      col_delim=col_delim, header_delim=header_delim)

        return pprint_dios(self, **kwargs)

    def to_df(self):
        df_or_ser = self._data.apply(lambda s: s).transpose()
        return pd.DataFrame() if isinstance(df_or_ser, pd.Series) else df_or_ser

    @property
    def debugDf(self):
        """ Return a aligned pd.DateFrame version of dios,
        where all columns of dtype obj/string.

        DO NOT USE FOR CALCULATIONS (!)"""
        return self.to_df()

    def apply(self, func, axis=0, raw=False, args=(), **kwds):
        if axis in [1, 'columns']:
            raise NotImplementedError

        elif axis in [0, 'index']:
            # we cannot use self._data.apply(func=func, args=args, **kwds)
            # because this may return a df
            need_dios = False
            new = []
            for c in self.columns:
                s = func(self._data.at[c].values if raw else self._data.at[c], *args, **kwds)
                new.append(s)
                if not _is_scalar(s):
                    need_dios = True

            if need_dios:
                data = pd.Series(dtype='O', index=self.columns)
                for i, c in enumerate(self.columns):
                    d = new[i]
                    if not isinstance(d, pd.Series):
                        d = pd.Series(d)
                    data.at[c] = d
                itype = _find_least_common_itype(data)
                new = DictOfSeries(data=data, itype=itype, fastpath=True)
            else:
                new = pd.Series(data=new, index=self.columns)
        else:
            raise ValueError(axis)
        return new

    def _fast_apply(self, attr, **kwargs):
        data = pd.Series(dtype='O', index=self.columns)
        call = callable(getattr(pd.Series, attr))
        for c in self.columns:
            s = self._data.at[c]
            data.at[c] = getattr(s, attr)(**kwargs) if call else getattr(s, attr)
        return data

    @property
    def dtypes(self):
        return self._fast_apply('dtype')

    def astype(self, dtype, copy=True, errors='raise'):
        data = self._fast_apply('astype', dtype=dtype, copy=copy, errors=errors)
        return DictOfSeries(data=data, itype=self.itype, cast_policy=self._policy, fastpath=True)

    def isna(self):
        data = self._fast_apply('isna')
        return DictOfSeries(data=data, itype=self.itype, cast_policy=self._policy, fastpath=True)

    def notna(self):
        data = self._fast_apply('notna')
        return DictOfSeries(data=data, itype=self.itype, cast_policy=self._policy, fastpath=True)

    def dropna(self, inplace=False):
        data = self._fast_apply('dropna', inplace=False)  # never pass inplace=True
        if inplace:
            self._data = data
        else:
            return DictOfSeries(data=data, itype=self.itype, cast_policy=self._policy, fastpath=True)

    def isin(self, values):
        data = self._fast_apply('isin', values=values)
        return DictOfSeries(data=data, itype=self.itype, cast_policy=self._policy, fastpath=True)

    def memory_usage(self, index=True, deep=False):
        mem = 0
        for k in self.columns:
            mem += self[k].memory_usage(index=index, deep=deep)
        return mem

    # #############################################################################
    # Operators

    def _op1(self, op):
        new = self.copy_empty(columns=True)
        try:
            for k in self.columns:
                new[k] = op(self[k])
        except Exception as e:
            raise type(e)(f"'{_OP_MAP[op]} dios' failed: " + str(e)) from e
        return new

    def _op2(self, op, other, align=True, inplace=False):
        def raiseif(kself, kother, s):
            if kself != kother:
                raise ValueError(f"{s} does not match, {s} left: {kself}, {s} right: {kother}")

        def doalign(left, right):
            return left.align(right, join='inner') if align else (left, right)

        def gen():
            if _is_dios_like(other):
                raiseif(list(self), list(other), 'keys')
                for k in self.columns:
                    left, right = self[k], other[k]
                    yield k, op(*doalign(left, right))
            elif isinstance(other, pd.Series):
                for k in self.columns:
                    left, right = self[k], other
                    yield k, op(*doalign(left, right))
            elif _is_dict_like(other):
                raiseif(sorted(self), sorted(other), 'keys')
                for k in self.columns:
                    yield k, op(self[k], other[k])
            elif _is_nested_list_like(other):
                raiseif(len(self), len(other), 'length')
                for i, k in enumerate(self.columns):
                    yield k, op(self[k], other[i])
            elif _is_scalar(other) or _is_list_like(other):
                for k in self.columns:
                    yield k, op(self[k], other)
            else:
                raise NotImplementedError

        new = self if inplace else self.copy_empty(columns=True)
        try:
            for k, val in gen():
                new[k] = val
        except Exception as e:
            raise type(e)(f"'dios {_OP_MAP[op]} other' failed: " + str(e)) from e
        return new

    __neg__ = ftools.partialmethod(_op1, op.neg)
    __abs__ = ftools.partialmethod(_op1, op.abs)
    __invert__ = ftools.partialmethod(_op1, op.inv)
    __eq__ = ftools.partialmethod(_op2, op.eq, align=False)
    __ne__ = ftools.partialmethod(_op2, op.ne, align=False)
    __le__ = ftools.partialmethod(_op2, op.le, align=False)
    __ge__ = ftools.partialmethod(_op2, op.ge, align=False)
    __lt__ = ftools.partialmethod(_op2, op.lt, align=False)
    __gt__ = ftools.partialmethod(_op2, op.gt, align=False)
    __add__ = ftools.partialmethod(_op2, op.add)
    __sub__ = ftools.partialmethod(_op2, op.sub)
    __mul__ = ftools.partialmethod(_op2, op.mul)
    __mod__ = ftools.partialmethod(_op2, op.mod)
    __truediv__ = ftools.partialmethod(_op2, op.truediv)
    __floordiv__ = ftools.partialmethod(_op2, op.floordiv)
    __pow__ = ftools.partialmethod(_op2, op.pow)
    __and__ = ftools.partialmethod(_op2, op.and_)
    __or__ = ftools.partialmethod(_op2, op.or_)
    __xor__ = ftools.partialmethod(_op2, op.xor)

    # dict-like methods

    def clear(self):
        d = self._data
        self._data = pd.Series(dtype=d.dtype, index=type(d.index)([]))
        
    def get(self, key, default=None):
        return self._data.get(key, default)

    def items(self):
        return self._data.items()

    def keys(self):
        return self.columns

    def pop(self, *args):
        # We support a default value, like dict, in contrary to pandas.
        # Therefore we need to handle args manually, because dict-style pop()
        # differ between a single arg and a tuple-arg, with arg and default,
        # where the second arg can be anything, including None. If the key is
        # not present, and a single arg is given, a KeyError is raised, but
        # with a given default value, it is returned instead.
        if len(args) == 0:
            raise TypeError("pop expected at least 1 arguments, got 0")
        if len(args) > 2:
            raise TypeError(f"pop expected at most 2 arguments, got {len(args)}")
        key, *rest = args
        if key in self.columns:
            return self._data.pop(key)
        elif rest:
            return rest.pop()
        raise KeyError(key)

    def popitem(self):
        last = self.columns[-1]
        return last, self._data.pop(last)

    def setdefault(self, key, default=None):
        if key not in self.columns:
            self._insert(key, default)
        return self._data[key]

    def update(self, other):
        if not _is_dios_like(other):
            other = to_dios(other)
        self.aloc[other, ...] = other


def _empty_repr(di):
    return f"Empty DictOfSeries\n" \
           f"Columns: {di.columns.to_list()}"


def pprint_dios(dios,
                max_rows=None, min_rows=None, max_cols=None,
                na_rep="NaN", empty_series_rep='no data',
                col_space=None, show_dimensions=True,
                col_delim=' | ', header_delim='='
                ):
    na_rep = str(na_rep)
    empty_series_rep = str(empty_series_rep)
    col_delim = col_delim or ' '

    min_rows = min(max_rows, min_rows)

    if dios.empty:
        return _empty_repr(dios)

    maxlen = dios.lengths.max()
    data = dios._data

    trunc_cols = len(data) > max_cols
    if trunc_cols:
        left, right = data.head(max_cols // 2), data.tail(max_cols // 2)
        data = left.append(right)

    # now data only contains series that we want to print.

    # if any series exceed max_rows we trim all series to min_rows
    series_lengths = data.apply(len).to_list()
    series_maxlen = max(series_lengths)
    trunc_rows = series_maxlen > max_rows
    max_rows = min_rows if trunc_rows else series_maxlen

    # we make a list of list, where the inner contains all
    # stringified values of the series upto max_rows+1, where
    # the additional row is the column-name
    outer = []
    for colname in data.index:
        s = data.at[colname]

        isempty = s.empty
        if isempty:
            s = pd.Series(empty_series_rep)
            idx = False
            cspace = col_space
        else:
            idx = True
            cspace = col_space // 2 if col_space else col_space

        sstr = s.to_frame().to_string(col_space=cspace,
                                      header=[str(colname)],
                                      index=idx,
                                      na_rep=na_rep,
                                      max_rows=max_rows,
                                      min_rows=min_rows,
                                      )
        li = sstr.split('\n')

        # HACK: empty series produce a unnecessary space,
        # because index is omitted
        if isempty:
            cstr, vstr = li
            if len(cstr.lstrip()) < len(vstr) and (cspace or 0) < len(vstr):
                li = [cstr[1:], vstr[1:]]

        outer.append(li)

    # now the length of every value-string per series are the same.
    # we need this length's to know, how many chars we need to fill,
    # once we exceed the length of the series, or if we insert whole
    # columns.
    valstr_len = [len(c[0]) for c in outer]

    rows = max_rows + 1  # colnames aka. header
    rows += 1 if trunc_rows else 0  # `...` in rows
    rows += 1 if header_delim else 0  # underline header

    if header_delim:
        for i, c in enumerate(outer):
            colheader = (header_delim * valstr_len[i])[:valstr_len[i]]
            c.insert(1, colheader)

    dots = ' ... '
    if trunc_cols:
        outer.insert(max_cols // 2, [dots] * rows)
        valstr_len.insert(max_cols // 2, len(dots))
        series_lengths.insert(max_cols // 2, rows)

    txt = ""
    for r in range(rows):
        for i, c in enumerate(outer):
            try:
                vstr = c[r]
            except IndexError:
                vstr = ' ' * valstr_len[i]
            txt += vstr + col_delim
        txt += '\n'

    # add footer
    if show_dimensions:
        for i, c in enumerate(outer):
            # ignore the dot-column
            if trunc_cols and i == max_cols // 2:
                txt += dots + ' ' * len(col_delim)
            else:
                txt += f"[{series_lengths[i]}]".ljust(valstr_len[i] + len(col_delim))

        txt += f'\n\nmax: [{maxlen} rows x {len(dios.columns)} columns]'
        txt += '\n'

    return txt


def _to_aligned_df(dios, no_value=' '):
    if dios.empty:
        return pd.DataFrame(columns=dios.columns)

    # keep track of all real nans
    nandict = {}
    for c in dios:
        nans = dios[c].isna()
        nandict[c] = nans[nans].index

    df = dios.to_df()
    df[df.isna()] = no_value

    # reinsert all real nans
    for c in df:
        df.loc[nandict[c], c] = np.nan

    return df


def _is_list_like_not_nested(obj):
    return _is_list_like(obj) and not _is_nested_list_like(obj)


def _is_dios_like(obj) -> bool:
    # must have columns
    # columns is some kind of pd.Index
    # iter will iter through columns
    # a `in` obj check if obj is in columns
    # obj[key] will give a pd.Series
    # obj.squeeze() give pd.Series if len(obj) == 1
    return isinstance(obj, DictOfSeries) or isinstance(obj, pd.DataFrame)


def _is_bool_series(obj) -> bool:
    return isinstance(obj, pd.Series) and obj.dtype == bool


def _is_bool_dios_like(obj) -> bool:
    if not _is_dios_like(obj):
        return False
    dtypes = obj.dtypes
    if (dtypes == bool).all():
        return True
    if (dtypes == 'O').any():
        return obj.apply(_is_bool_indexer).all()
    return False


def to_dios(obj) -> DictOfSeries:
    if isinstance(obj, DictOfSeries):
        return obj
    return DictOfSeries(data=obj)


def __monkey_patch_pandas():
    def to_dios(self):
        return DictOfSeries(data=self)

    pd.Series.to_dios = to_dios
    pd.DataFrame.to_dios = to_dios


__monkey_patch_pandas()

# keep this here to prevent cyclic import
from .indexer import _aLocIndexer, _iLocIndexer, _LocIndexer, _iAtIndexer, _AtIndexer
